# SPDX-License-Identifier: Apache-2.0

# This file defines the interface of the converter internal object for callback,
# So the usage of the methods and properties list here will not be affected among the different versions.

from onnxconverter_common.interface import *  # noqa
